from typing import List
from napalm.eos import EOSDriver
from ..base import SCBaseDriver
from ..models import InventoryDict


class SCEOSDriver(EOSDriver, SCBaseDriver):
    ("fabric_module",)
    ("fan",)
    ("linecard",)
    ("optic",)
    ("psu",)
    ("re",)
    ("stack_cable",)
    ("stack_member",)
    ("uplink_module",)
    ("aoc",)
    ("dac",)

    INVENTORY_TO_TYPE = {
        r"Fabric": "fabric_module",
        r"Linecard": "linecard",
        r"Supervisor": "re",
    }

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Forcing ssh transport since we don't enable the web interface
        """
        optional_args = optional_args if optional_args else {}
        optional_args["transport"] = "ssh"

        super().__init__(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

    def get_inventory(self) -> List[InventoryDict]:
        inventory = self._run_commands(["show inventory"], encoding="json")

        results = []

        ### optics
        for slot, optic in inventory[0]["xcvrSlots"].items():
            if optic.get("modelName"):
                results.append(
                    {
                        "type": "optic",
                        "subtype": optic["modelName"],
                        "name": f"Ethernet{slot}",
                        "part_number": optic["modelName"],
                        "serial_number": optic["serialNum"],
                    },
                )

        ### line cards
        for slot, card in inventory[0]["cardSlots"].items():
            if card.get("serialNum"):
                results.append(
                    {
                        "type": self._get_inventory_type(slot),
                        "subtype": card["modelName"],
                        "name": f"Ethernet{slot}",
                        "part_number": card["modelName"],
                        "serial_number": card["serialNum"],
                    },
                )

        ### PSUs
        for slot, psu in inventory[0]["powerSupplySlots"].items():
            if psu.get("serialNum"):
                results.append(
                    {
                        "type": "psu",
                        "subtype": None,
                        "name": f"PSU {slot}",
                        "part_number": psu["name"],
                        "serial_number": psu["serialNum"],
                    },
                )

        ### FANs
        for slot, fan in inventory[0]["fanTraySlots"].items():
            if fan.get("serialNum"):
                results.append(
                    {
                        "type": "fan",
                        "subtype": None,
                        "name": f"FAN {slot}",
                        "part_number": fan["name"],
                        "serial_number": fan["serialNum"],
                    },
                )
        return results



GET_INVENTORY = {
    "_": """
    <filter xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
        <state xmlns="urn:nokia.com:sros:ns:yang:sr:state">
            <port>
                <port-id/>
                <transceiver>
                    <vendor-part-number/>
                    <model-number/>
                    <vendor-serial-number/>
                    <connector-type/>
                    <type/>
                </transceiver>
            </port>
        </state>
    </filter>
    """
}



import argparse
from pprint import pprint

from napalm import get_network_driver

from custom_napalm import PLATFORM_MAP
from custom_napalm.base import SCBaseDriver
from custom_napalm.utils import configure_logging, LOG_LEVELS, get_from_args_or_env

# list of getters to run
GETTERS = [attr for attr in SCBaseDriver.__dict__ if attr.startswith("get")]

cred_args = {"sc_username": True, "sc_password": True}


def main():
    parser = argparse.ArgumentParser(
        description="""
Run a specific sc_napalm "getter" against a device.
"""
    )
    parser.add_argument("device", help="device hostname or IP address")
    parser.add_argument(
        "sc_napalm_platform",
        choices=PLATFORM_MAP,
        help="The platform of this device",
    )
    parser.add_argument(
        "cmd",
        choices=GETTERS,
        help="The getter command to run against this device",
    )
    parser.add_argument("--ssh-cfg", help="Use SSH config file to connect", type=str)
    log_args = parser.add_mutually_exclusive_group()
    log_args.add_argument(
        "-l", "--log-level", help="Set log level for sc_napalm only", choices=LOG_LEVELS
    )
    log_args.add_argument(
        "-L", "--LOG-LEVEL", help="set global log level", choices=LOG_LEVELS
    )
    parser.add_argument(
        "--logfile",
        type=str,
        help="Save logging to a file (specified by name) instead of to stdout",
    )

    for cred_arg in cred_args:
        parser.add_argument(f"--{cred_arg}", help="Specify credentials")
    args = parser.parse_args()

    log_level = args.log_level if args.log_level else args.LOG_LEVEL
    if log_level:
        configure_logging(
            log_level,
            log_globally=bool(args.LOG_LEVEL),
            log_file=args.logfile,
            log_to_console=not (bool(args.logfile)),
        )

    creds = {
        cred: get_from_args_or_env(cred, args, required=reqd)
        for cred, reqd in cred_args.items()
    }
    Driver = get_network_driver(args.sc_napalm_platform)

    # setting up connection details
    optional_args = {"look_for_keys": False}
    if args.ssh_cfg:
        optional_args = {"ssh_config_file": args.ssh_cfg}

    with Driver(
        args.device,
        creds["sc_username"],
        creds["sc_password"],
        timeout=60,
        optional_args=optional_args,
    ) as conn:
        result = getattr(conn, args.cmd)()
        pprint(result)


if __name__ == "__main__":
    main()



## https://www.juniper.net/documentation/us/en/software/junos-pyez/junos-pyez-developer/topics/task/junos-pyez-tables-views-loading.html

"""
Load tables/views
"""

from os.path import splitext
import yaml
from jnpr.junos.factory import FactoryLoader


_YAML_ = splitext(__file__)[0] + ".yml"

with open(_YAML_, encoding="utf-8") as f:
    yaml_str = f.read()

globals().update(FactoryLoader().load(yaml.safe_load(yaml_str)))



# # pylint: disable=no-member
from typing import List
import re
import logging

from napalm.junos import JunOSDriver
from napalm.base.models import LLDPNeighborDict, ConfigDict
from ..models import InventoryDict
from ..base import SCBaseNetconfDriver


from . import junos_views

logger = logging.getLogger(__name__)

XCVR_DESC_TO_INTERFACE_PREFIX = {
    r"100GBASE-": "et-",
    r"QSFP": "et-",
    r"SFP28": "et-",
    r"SFP\+": "xe-",
    r"SFP-": "ge-",
    r"CFP-": "et-",
    r"(UNKNOWN|UNSUPPORTED)": "ge-",  # default
}

DEFAULT_ARP_AGING_TIMER = 1200


class SCJunOS(JunOSDriver, SCBaseNetconfDriver):
    """
    Junos parser
    """

    netmiko_host_type = "juniper_junos"

    # mapping of inventory name regex to their netbox types
    INVENTORY_TO_TYPE = {
        r"Fan Tray": "fan",
        r"(PIC|MIC|FPC)": "linecard",
        r"Xcvr": "optic",
        r"(PEM|Power Supply)": "psu",
        r"Routing Engine": "re",
        # don't care about these inventory items
        r"(CPU|PDM|FPM|Midplane|CB|SFB|TIB)": None,
    }

    def _get_junos_inventory_type(self, name: str, model_number: str) -> str:
        """
        Maps the name and part number of the Junos inventory item to its type
        """
        # "Xcvr" parts are always optics
        if name.startswith("Xcvr"):
            return "optic"

        # non-optic parts should always have a model number, if not
        # we don't care about them
        if not (model_number):
            return None

        # EX switches and virtual chassis save their vc members and uplink modules
        # as FPC X and PIC X - we want to classify those correctly
        if re.match(r"EX[234]", model_number):
            if re.search(r"FPC \d+", name):
                return "stack-member"

            if re.search(r"PIC \d+", name):
                return None

        # uplink modules are also saved under PIC X, must classify based
        # on their model number
        if re.match(r"EX-UM", model_number):
            return "uplink-module"

        # otherwise we want to pattern-match based on the INVENTORY_TO_TYPE
        # dictionary
        return self._get_inventory_type(name)

    def _get_xcvr_interface_prefix(self, xcvr_desc: str) -> str:
        """
        Maps xcvr description from "show chassis hardware" to
        interface prefix
        """
        for pattern, prefix in XCVR_DESC_TO_INTERFACE_PREFIX.items():
            if re.match(pattern, xcvr_desc):
                return prefix

        raise ValueError(f"{self.hostname}: Unknown xcvr type {xcvr_desc}")

    def _get_mod_number(self, mod_name: str) -> str:
        """
        Strips out "FPC|MIC|PIC" from a module name seen in "show chassis hardware"
        """
        return re.sub(r"(MIC|PIC|FPC) ", "", mod_name)

    def _get_inventory_part_number(self, item: any) -> str:
        """
        Extracts the part number from an inventory item
        which, depending on the specific item, is stashed in
        the part number, model number, or description field
        """
        if item.model_number and item.model_number != "model-number":
            return item.model_number
        return item.part_number

    def _save_inventory_item(
        self, output: list, item: any, parent: str = None, grandparent: str = None
    ) -> bool:
        """
        Extracts data from a particular inventory item object.
        Returns whether we care aobut this item or not (so we know whether or not)
        to loop over its children
        """
        # skip builtin types, or parts without model numbers that aren't transcievers
        if item.part_number == "BUILTIN":
            return False

        # get inventory type based on item name and P/N
        inv_type = self._get_junos_inventory_type(item.name, item.model_number)

        if not inv_type:
            return False

        # for transcievers, change the name from Xcvr X to be the junos interface name
        # note that we expect transcievers to always be at the sub-sub module level
        # and thus to have a grandparent and parent
        m = re.search(r"Xcvr (\d+)", item.name)
        if m:
            if not (grandparent or parent):
                raise ValueError(
                    f"{self.hostname}: No MIC and PIC found for {item.name}"
                )
            prefix = self._get_xcvr_interface_prefix(item.description)

            item_name = f"{prefix}{self._get_mod_number(grandparent)}/{self._get_mod_number(parent)}/{m.group(1)}"

        # for sub-linecards (mics or pics) we want to prepend the parent to the item name
        elif parent:
            item_name = f"{parent} {item.name}"
        else:
            item_name = item.name

        output.append(
            {
                "type": inv_type,
                "name": item_name,
                "subtype": item.description,
                "part_number": self._get_inventory_part_number(item),
                "serial_number": item.serial_number,
            }
        )

        return True

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> ConfigDict:
        """
        pulling running config via ssh sow we can have it in 'display set' format.
        """
        if format != "text":
            return super().get_config(
                retrieve=retrieve, full=full, sanitized=sanitized, format=format
            )

        config = {"startup": "", "running": "", "candidate": ""}

        with self.ssh_conn() as ssh_device:
            config["running"] = ssh_device.send_command(
                "show configuration | display set"
            )

        return config

    def get_lldp_neighbors(self) -> LLDPNeighborDict:
        """
        Overrides napalm get_lldp_neighbors so that we actually get the remote
        interface ID instead of the description.

        NAPALM's get_lldp_neighbors_detail gets this for us - note that to get this detail,
        you need to execute a command for each interface just like on the cli :(
        """
        neighs = super().get_lldp_neighbors_detail()
        output = {}
        for port, neighs in neighs.items():
            output[port] = []
            for neigh in neighs:
                output[port].append(
                    {
                        "hostname": neigh["remote_system_name"]
                        if neigh["remote_system_name"]
                        else neigh["remote_chassis_id"],
                        "port": neigh["remote_port"],
                    }
                )
        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        parses the get-chassis-inventory RPC call, which maps to "show chassis hardware"
        """
        result = junos_views.junos_inventory(self.device).get()

        output = []
        for chassis in dict(result).values():
            # note that the chassis model and s/n are at this level, but
            # that doesn't count as an 'inventory item' so we're ignoring it

            # saving modules, sub-modules and sub-sub modules
            for module in dict(chassis.modules).values():
                self._save_inventory_item(output, module)

                for sub_module in dict(module.sub_modules).values():
                    self._save_inventory_item(output, sub_module, parent=module.name)

                    for sub_sub_module in dict(sub_module.sub_sub_modules).values():
                        self._save_inventory_item(
                            output,
                            sub_sub_module,
                            parent=sub_module.name,
                            grandparent=module.name,
                        )

        return output



from typing import Dict, List
import urllib3
from napalm.base import models as napalm_models, NetworkDriver

import math
import re
import ipaddress

from pyaoscx.session import Session
from pprint import pprint
from ..base import SCBaseNetconfDriver
from .. import models

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DEFAULT_VERSION = "10.04"


class SCAOSCXDriver(NetworkDriver, SCBaseNetconfDriver):
    """
    Current scaos napalm driver is old and only supports api v1, which is
    deprecated. So we are writing our own one. Woot!
    """

    netmiko_host_type = "aruba_aoscx"

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """NAPALM Constructor for AOS-CX."""
        if optional_args is None:
            optional_args = {}

        if "version" not in optional_args:
            optional_args["version"] = DEFAULT_VERSION
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.platform = "aoscx"

        # Hack to get IPv6 rest calls to work - need the IP address to be encased
        # in brackets
        try:
            ipaddress.IPv6Address(hostname)
            hostname = f"[{hostname}]"
        except ValueError:
            pass

        self.session = Session(hostname, optional_args["version"])

    def open(self):
        """
        Implementation of NAPALM method 'open' to open a connection to the device.
        """
        self.session.open(self.username, self.password)

    def close(self):
        """
        Implementation of NAPALM method 'close'. Closes the connection to the device and does
        the necessary cleanup.
        """
        self.session.close()

    def _get(self, path: str, attrs: list = None, depth: int = 1):
        path = re.sub(r"/rest/v.+?/", "", path)
        params = None
        if attrs:
            params = {"attributes": ",".join(attrs), "depth": depth}
        r = self.session.request("GET", path, params=params)
        r.raise_for_status()
        return r.json()

    def _iter_get(
        self, path: str, sub_path: str = None, attrs=None, depth=1
    ) -> dict[str, any]:
        """
        Iterate through a dict of paths at 'path', doing rest 'gets' for the
        a particular sub_path on each one with the specificed attributes and depth
        :path: base path, eg 'system/interfaces'
        :sub_path: append to the base path, eg 'lldp_neighbors'
        :attrs: restrict output at each sub_path

        returns dict of data mapping to each key in the original path
        """
        results = {}
        paths = self._get(path, depth=1)
        for k, v in paths.items():
            full_path = f"{v}/{sub_path}" if sub_path else v
            results[k] = self._get(full_path, attrs, depth=depth)

        return results

    def json_lookup(self, data, key_name, default_value="", one_match=True):
        result = []
        self._json_lookup(data, key_name, result, one_match=one_match)

        if not result:
            return default_value
        if one_match:
            return result[0]
        return result

    def _json_lookup(self, data, key_name, result, one_match):
        """
        recurses through a nested dictionary/list structure till
        it finds a key in a dict, and returns a value.
        If we have to iterate through a list along the way we will
        recurse down each item on the list.
        """
        if one_match and len(result) != 0:
            return

        if isinstance(data, list):
            for entry in data:
                self._json_lookup(entry, key_name, result, one_match)

        elif isinstance(data, dict):
            for k, v in data.items():
                if key_name == k:
                    result.append(data[key_name])
                else:
                    self._json_lookup(v, key_name, result, one_match)

    def get_facts(self) -> napalm_models.FactsDict:
        sys_attrs = [
            "software_version",
            "boot_time",
            "applied_domain_name",
            "applied_hostname",
        ]
        sys_data = self._get("system", attrs=sys_attrs)
        chassis = self._get("system/subsystems/chassis,1", attrs=["product_info"])

        facts = {
            "os_version": sys_data.get("software_version", ""),
            "uptime": sys_data.get("boot_time", 0),
            "interface_list": list(self._get("system/interfaces")),
            "vendor": "Aruba",
            "serial_number": self.json_lookup(chassis, "serial_number"),
            "model": self.json_lookup(chassis, "part_number"),
            "hostname": sys_data.get("applied_hostname", ""),
            "fqdn": sys_data.get("applied_domain_name", ""),
        }

        return facts

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        if sanitized:
            raise NotImplementedError("Sanitized nos config not supported")
        if format not in ["text", "json"]:
            raise NotImplementedError("Only text and json config formats supported")

        # for text format we're going to SSH into the device and do 'show config'
        # in this scenario 'candidate' isn't really relevant.
        config = {"startup": "", "running": "", "candidate": ""}
        if format == "text":
            with self.ssh_conn() as ssh_device:
                config["running"] = ssh_device.send_command("show running-config")
                config["startup"] = ssh_device.send_command("show startup-config")
        else:
            config["running"] = self._get("fullconfigs/running-config")
            config["startup"] = self._get("fullconfigs/startup-config")

        return config

    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        output = {}
        i_lldp = self._iter_get("system/interfaces", sub_path="lldp_neighbors")

        # lldp endpoint for each interface is itself a path that must be resolved
        # to get more than just the mac address of each neighbor
        for port, neighbors in i_lldp.items():
            for n_path in neighbors.values():
                n_details = self._get(n_path)

                if not output.get(port):
                    output[port] = []

                output[port].append(
                    {
                        "hostname": self.json_lookup(n_details, "chassis_name"),
                        "port": n_details["port_id"],
                    }
                )

        return output
    
    def get_optics(self) -> Dict[str, napalm_models.OpticsDict]:

        with self.ssh_conn() as ssh_device:
            result = ssh_device.send_command("show interface dom")

            output = {}
            for line in result.split("\n"):

                m = re.match(r'(\d+\/\d+\/\d+)\s+\S+\s+\S+\s+\S+\s+([\d\.\-]+)\s+([\d\.\-]+)\s+([\d\.\-]+)', line)
                if m:
                    i_name = m.group(1)
                    output[i_name] = {"physical_channels": {"channels": {}}}
                    i_channels = output[i_name]["physical_channels"]["channels"]
                    i_channels[0] = {
                        "laser_bias_current": m.group(2),
                        "input_power": m.group(3),
                        "output_power": m.group(4),
                    }

            return output





    # def get_optics(self) -> Dict[str, napalm_models.OpticsDict]:
    #     output = {}
    #     interfaces = self._iter_get("system/interfaces", attrs=["pm_monitor"])

    #     for i_name, data in interfaces.items():
    #         i_data = data["pm_monitor"]

    #         if not i_data:
    #             continue

    #         output[i_name] = {"physical_channels": {"channels": {}}}
    #         i_channels = output[i_name]["physical_channels"]["channels"]
    #         for channel, c_data in i_data.items():
    #             if channel == "common":
    #                 continue
    #             i_channels[channel] = {
    #              "input_power": 10 * math.log(c_data["rx_power"]),
    #              "output_power": 10 * math.log(c_data["tx_power"]),
    #              "laser_bias_current": c_data.get("tx_bias", "0.0"),
    #             }

    #     return output

    # def get_optics(self) -> Dict[str, napalm_models.OpticsDict]:
    #     output = {}
    #     interfaces = self._iter_get("system/interfaces", attrs=["pm_info"])
    #     for i_name, data in interfaces.items():
    #         i_data = data["pm_info"]
            
    #         # if there's no optic installed or optics aren't supported
    #         # skip this port
    #         if i_data.get("connector", "absent") == "absent":
    #             continue

    #         output[i_name] = {"physical_channels": {"channels": {}}}
    #         i_channels = output[i_name]["physical_channels"]["channels"]

    #         # single-channel optics - not sure how multi-channel optics
    #         # look yet
    #         if i_data.get("rx_power"):
    #             i_channels["0"] = {
    #                 "input_power": 10 * math.log(i_data["rx_power"]),
    #                 "output_power": 10 * math.log(i_data["tx_power"]),
    #                 "laser_bias_current": i_data.get("tx_bias", "0.0"),
    #             }

    #     return output

    def get_inventory(self) -> List[models.InventoryDict]:
        output = []
        optics = self._iter_get("system/interfaces", attrs=["pm_info"])

        for i_name, optic in optics.items():
            i_data = optic["pm_info"]

            # if there's no optic installed or optics aren't supported
            # skip this port
            if i_data.get("connector", "absent") == "absent":
                continue

            output.append(
                {
                    "type": "optic",
                    "subtype": i_data.get("xcvr_desc", ""),
                    "name": i_name,
                    "part_number": i_data.get("vendor_part_number", ""),
                    "serial_number": i_data.get("vendor_serial_number", ""),
                }
            )

        return output



from typing import List
import logging

from lxml import etree

from napalm.iosxr_netconf import IOSXRNETCONFDriver
from ncclient.xml_ import to_ele

from ..base import SCBaseNetconfDriver
from ..models import InventoryDict

from .constants import INV_RPC_REQ

logger = logging.getLogger(__name__)


class SCIOSXR(IOSXRNETCONFDriver, SCBaseNetconfDriver):
    """
    IOSXR Class
    """

    INVENTORY_TO_TYPE = {
        r"PM\d+$": "psu",
        r"FT\d+$": "fan",
        r"FC\d+$": "fabric_module",
        r"RP\d$": "re",
        r"SC\d": None,  # don't care about system controller
        r"Rack 0": None,  # this is the chassis, don't care about it
        r"^(FourHundred|Hundred|Forty)GigE": "optic",
        r"\d/\d": "linecard",
    }

    def get_inventory(self) -> List[InventoryDict]:
        """
        Gets inventory data
        """
        rpc_reply = self.device.dispatch(to_ele(INV_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)

        output = []
        for item in self._xpath(xml_result, "//inv:inv-basic-bag"):
            name = self._text(item, "inv:name")
            item_type = self._get_inventory_type(name)
            if item_type:
                output.append(
                    {
                        "type": item_type,
                        "name": name,
                        "part_number": self._text(item, "inv:model-name"),
                        "serial_number": self._text(item, "inv:serial-number"),
                        "parent": None,
                    }
                )

        return output



# from .iosxr_netconf import SCIOSXR
from .iosxr import SCIOSXR_SSH
from .iosxr_netconf import SCIOSXR
from .junos import SCJunOS
from .nxos import SCNXOS
from .sros import SCNokiaSROSDriver
from .srl_ssh import NokiaSRLSSHDriver
from .srl import SCNokiaSRLDriver
from .eos import SCEOSDriver
from .nos import SCNOSDriver
from .waveserver import WaveServerDriver
from .aoscx import SCAOSCXDriver

PLATFORM_MAP = {
    "iosxr": SCIOSXR_SSH,
    "iosxr_netconf": SCIOSXR,
    "nxos": SCNXOS,
    "junos": SCJunOS,
    "sros": SCNokiaSROSDriver,
    "srl_ssh": NokiaSRLSSHDriver,
    "srl": SCNokiaSRLDriver,
    "eos": SCEOSDriver,
    "nos": SCNOSDriver,
    "waveserver": WaveServerDriver,
    "aoscx": SCAOSCXDriver,
}


def get_network_driver(platform: str):
    """
    Returns network driver based on platform string.
    """
    for valid_platform, driver in PLATFORM_MAP.items():
        if valid_platform == platform:
            return driver

    raise NotImplementedError(f"Unsupported platform {platform}")
